(ns com.swir.linux.tm.luarpc.core
  (:use aleph.tcp lamina.core gloss.core gloss.io 
    [gloss.core.protocols :exclude (sizeof)]
    com.swir.linux.tm.luarpc.session))

(defmacro recur-codec [codec]
  `(reify
     gloss.core.protocols.Reader
     (gloss.core.protocols/read-bytes [this#  buf-seq#]
       (gloss.core.protocols/read-bytes ~codec buf-seq#))
     gloss.core.protocols.Writer
     (gloss.core.protocols/sizeof [this#] nil)
     (gloss.core.protocols/write-bytes [this# buf# val#]
       (gloss.core.protocols/write-bytes ~codec buf# val#))))

(defn si2us
  "Convert a java int (signed) to an unsigned short stored in a Java short (signed)" 
  [i]
  (if (> i 32767)
    (-> i bit-not (bit-and 0xFFFF) inc - byte)
    (byte i)))

(defn us2si 
  "Convert an unsigned short stored in a java short (signed) to a java int (signed)" 
  [b]
  (if (neg? b)
    (-> b - dec bit-not (bit-and 0xFFFF) int)
    (int b)))

(defrecord reference [i])
(def ref? (partial instance? reference))

(defrecord function [])
(def func? (partial instance? function))

(defn pre-encode-data [v]
  (cond
    (nil? v)
    [:nil]
    (instance? Boolean v)
    [:boolean (if v (byte 1) (byte 0))]
    (integer? v)
    [:integer v]
    (number? v)
    [:number v]
    (string? v)
    [:string v]
    (keyword? v)
    [:string (name v)]  
    (sequential? v)
    [:table (map-indexed #(vector (inc %1) %2) v)]
    (map? v)
    [:table (into [] v)]))

(defn post-decode-data [[t v]]
  (case t
    :nil
    nil
    :boolean
    (if (zero? v) false true)
    :integer
    (int v)
    :number
    (let [d v   i (int d)] (if (= i d) i d))
    :string
    (str v)
    :table
    (into [] v) 
    :ref
    (reference. (us2si (dec v)))))

(declare deref-elem) 

(defn- deref-tab [e b]
  (let [idx (count b)
        prom (promise)
        nb (conj b prom)
        [flat nb] (reduce #(let [[a b] %1 [ne nb] (deref-elem %2 b)] [(conj a ne) nb]) [[] nb] (mapcat identity e))
        m (apply array-map flat)
        c (take (count (keys m)) (rest (range)))
        ne (if (= c (keys m)) (into [] (vals m)) m)]
    (deliver prom ne)
    [ne (assoc nb idx ne)]))

(defn deref-elem
  ([e] (first (deref-elem e [])))
  ([e b]
    (cond
      (string? e) [e (conj b e)]
      (ref? e) [(get b (:i e)) b]
      (vector? e) (deref-tab e b) 
      :default [e b]))) 



(declare data)
(defcodec data-type (enum :byte {:nil 0 :number 3 :boolean 1 :string 4 :table 5 :function 6 :integer 7 :ref 0x14}))
(defcodec data-ni [:nil])
(defcodec data-bo [:boolean :byte])
(defcodec data-nb [:number :float64])
(defcodec data-st [:string (finite-frame :int16 (string :ascii))])
(defcodec data-tb [:table (repeated [(recur-codec data) (recur-codec data)] :prefix :int16)])
(defcodec data-rf [:ref :int16])
(defcodec data-in [:integer :int32])
(def data (compile-frame
            (header data-type
              {:nil data-ni :boolean data-bo :number data-nb :string data-st :table data-tb :ref data-rf :integer data-in}
              first)
            pre-encode-data post-decode-data))
(defcodec command-type (enum :byte {:send 0, :resp 1}))
(defcodec command-send [:send :byte (finite-frame :int32 [data (repeated data :prefix :none)])])
(defcodec command-resp [:resp :byte (finite-frame :int32 [:byte (repeated data :prefix :none)])])
(defcodec command (header command-type {:send command-send :resp command-resp} first))

(defn- notify-clbk [clbk value]
  (cond 
    (fn? clbk) (clbk value)
    (nil? clbk) nil
    :default (deliver clbk value)))

(declare stop-client)

(defn- receive-command [ch token cmd]
   (let [[t i [a b]] cmd
         r (map deref-elem b)]
     (case t
       :resp (remove-clbk token i #(notify-clbk % r))
       :send (let [resp (call-resp-clbk token a r)
                   r [:resp i [(byte 0) (or resp [])]]]
               (enqueue ch r)))))

(defn start-client [host port]
  (let [ch (wait-for-result (tcp-client {:host host :port port :frame command}))
        token (add-session (str host ":" port) ch)]
    (receive-all ch (partial receive-command ch token))
    token))

(defn stop-client [token]
  (remove-session token #(notify-clbk % :closed) #(when % (close %))))

(defn- timer [token id timeout]
  (future
    (Thread/sleep timeout)
    (remove-clbk token id #(notify-clbk % :timeout))))

(defn cmd [token timeout cmd-name & cmd-args]
  (let [ch (@token-session token)
        p (promise)
        id (add-clbk token p)
        c [:send id [cmd-name (or cmd-args [])]]]
    (when timeout (timer token id timeout))
    (enqueue ch c)
    @p))

(defn cmd-clbk [token timeout clbk cmd-name & cmd-args]
  (let [ch (@token-session token)
        id (add-clbk token clbk)
        c [:send id [cmd-name (or cmd-args [])]]]
    (when timeout (timer token id timeout))
    (enqueue ch c))) 

(comment
(def servers (atom {}))

(defn add-cmd-handler [cmd-name handler-fn]
  (swap! servers assoc cmd-name handler-fn))

(defn remove-cmd-handler [cmd-name]
  (swap! servers dissoc cmd-name)))

